#ifndef UNITYLIB_H_LOADED
#define UNITYLIB_H_LOADED 1

/**
 * @file unity.h
 */

/* forward definition -- see unit-definitions.h */
struct unit_definition;

/* forward definition -- see function-definitions.h */
struct function_definition;

/** @struct FunctionApplication
 * A unit corresponding to an application of a function to a unit
 * sequence.  In the case of 'log(V^2/Hz)', for example, the function
 * is 'log' and the operand is the parsed units corresponding to
 * 'V^2/Hz'.
 */
typedef struct {
    /** The definition of the function used.
     * Precisely one of definition and name should be non-null */
    const struct function_definition* definition;
    /** The name of an unreognised function used.
     * Precisely one of definition and name should be non-null */
    const char* name;
    /** The parsed units to which the function is applied. */
    const struct unit_struct* operand;
} FunctionApplication;

/** @struct SimpleUnit
 * A simple unit, corresponding to, for example, 'MHz'.
 */
typedef struct {
    /**
     * The base-10 logarithm of the prefix in front of the unit.  For
     * example, this will be '-3' for the 'm' (milli-) prefix.
     */
    int prefix_power;           /* -24 to +24 */
    /**
     * True (non-zero) if the prefix is a power of 10;
     * false if it's a power of 2.
     */
    int base10_p;
    /**
     * The base unit of this unit specification, as a known unit.  If
     * this is not a known unit (in the syntax which parsed the
     * associated expression) then this will be NULL.
     */
    const struct unit_definition* base_unit_def;
    /**
     * The base unit of this unit specification.  This might be 'm'
     * for metres, 'Hz' for Hertz, and so on.
     */
    char* base_unit_string;     /* eg Hz or aangstrom */
} SimpleUnit;

/** An enumeration indicating the two types of unit possible in
 *  {@link unit_struct} */
typedef enum {
    /** Simple unit type (eg 'MHz') */
    simple_unit_type,
    /** Function-of unit type (eg 'log(MHz)') */
    function_application_type
} UnitTypes;

/** @struct unit_struct
 * A single parsed unit.  This can be either a simple unit
 * (corresponding to, for example, 'MHz') or a function of a sequence
 * of units (corresponding to, for example, 'log(V^2/mHz)').
 */
struct unit_struct {
    /**
     * The power to which the unit is raised.  The specification
     * 'mm^(-2)' will result in a power of '-2'.
     */
    float power;
    /**
     * Indicates which type of unit this is.
     * Depending on the value of this, one or other of the structs 'u'
     * or 'f' will be valid (but not both).
     */
    UnitTypes type;
    union {
        /** A contained simple unit */
        SimpleUnit u;
        /** A contained function-of unit */
        FunctionApplication f;
    };
    /** A pointer to the next unit in the sequence */
    struct unit_struct *next;
};


/** A clearer name for struct unit_struct */
typedef struct unit_struct Unit;

/** @struct UnitExpression
 * The parsed version of a unit string.
 */
typedef struct {
    /**
     * The base-10 log of the numerical factor at the head of the string.
     * For some parsers, this is restricted to round powers of ten.
     */
    float log_factor;
    /**
     * The sequence of units which comprise the parsed units specification.
     */
    const Unit* unit_sequence;
} UnitExpression;

const UnitExpression* unity_parse_string(const char* unit_string, const int syntax);

/* The function unity_parse_error is declared (and documented) here,
   and defined in lex.lex */
/**
 * Retrieve the most recent lexing or parsing error.  Immediately after
 * the parser has returned with an error (that is function
 * {@link #unity_parse_string} has returned NULL) this function can be called
 * to obtain an explanation.  However, a non-null response from this
 * function does not imply that the most recent call to `unity_parse_string' failed.
 *
 * <p>The returned string should not be freed by the caller.
 * @return a static string containing an explanation of a parse error
 */
const char* unity_parse_error(void);

const char** unity_parser_names(void);

int unity_identify_parser(const char* parser_name);
const char* unity_parser_name(const int parser_id);
void unity_free_expression(const UnitExpression* u);
const char* unity_version_string();
int unity_version_number();

/* Checking units */
/** Flag: check whether the unit is recognised.  See {@link #unity_check_unit} */
#define UNITY_CHECK_RECOGNISED 1
/** Flag: check whether the unit is recommended.  See {@link #unity_check_unit} */
#define UNITY_CHECK_RECOMMENDED 2
/** Flag: check whether the unit is used in a way which satisfies constraints.  See {@link #unity_check_unit} */
#define UNITY_CHECK_CONSTRAINTS 4
/** Flag: perform all checks.
 * Do not rely on the numerical value of this or its related constants,
 * which may change between library versions without notice.
 * See {@link #unity_check_unit}
 */
#define UNITY_CHECK_ALL UNITY_CHECK_RECOGNISED | UNITY_CHECK_RECOMMENDED | UNITY_CHECK_CONSTRAINTS
int unity_check_unit(const Unit* u, const int syntax, const int flags);
int unity_check_expression(const UnitExpression* ue, const int syntax, const int flags);

#ifdef UNITY_INTERNAL
/* Internal functions */
Unit* u_new_unit(const char* unitString, const float power, int syntax);
Unit* u_function_application(const char* functionName,
                             const Unit* unit_sequence,
                             int syntax);
Unit* u_unit_append(Unit* u1, Unit* u2);
void u_receive_result(float, const Unit* unit_sequence);
int u_prefix_to_power(const char pfx);
//char* u_copy_string(const char c, const char* s);
Unit* u_divide_units(Unit* num, Unit* den);
Unit* u_unit_reciprocal(Unit* u);
void u_free_unit(const Unit* u);
#endif

#endif
