/*
 * This module performs the tests described in ../../grammar/testcases.csv. 
 *
 * Usage: java -cp ... ParseAll -t -pxxx ../../grammar/testcases-xxx.csv
 *
 * Here, 'xxx' is one of the supported parsers: fits, ogip or cds.
 */

import uk.me.nxg.unity.*;
import uk.me.nxg.unity.util.SimpleCSVReader;

import java.util.List;

public class ParseAll 
{
    public static void main(String[] args) 
    {
        final int LIST_LEXEMES = 1;
        final int DISPLAY_PARSE = 2;
        final int RUN_TESTS = 3;
        int action = 0;
        
        String inputString = null;
        char optionChar;
        String parserName = "fits";

        for (int i=0; i<args.length; i++) {
            if (args[i].charAt(0) == '-') {
                switch (args[i].charAt(1)) {
                  case 'l':
                    action = LIST_LEXEMES;
                    break;
                  case 'd':
                    action = DISPLAY_PARSE;
                    break;
                  case 't':
                    action = RUN_TESTS;
                    break;
                  case 'p':
                    parserName = args[i].substring(2);
                    break;
                  case 'S':
                    {
                        for (Syntax s : Syntax.values()) {
                            System.out.println("parser: " + s);
                        }
                    }
                    break;
                  default:
                    Usage();
                    break;
                }
            } else {
                inputString = args[i];
            }
        }

        if (inputString == null || action == 0) {
            Usage();
        }

        // if (! UnitParser.getParserTypes().contains(parser)) {
        //     System.err.println(""

        Syntax parser = Syntax.lookup(parserName);
        if (parser == null) {
            System.err.println("Unrecognised parser name: " + parserName);
            System.exit(1);
        }
        
        try {
            switch (action) {
              case LIST_LEXEMES:
                listLexemes(inputString);
                break;
              case DISPLAY_PARSE:
                // parse the input string and display the contents in an unambiguous way
                displayParse(inputString, parser);
                break;
              case RUN_TESTS:
                runTestsInFile(inputString, parser);
                break;
              default:
                Usage();
            }
        } catch (UnitParserException e) {
            System.err.println("Unit parser exception: " + e);
        }
    }

    /**
     * List lexemes in the inputString, for debugging 
     */
    private static void listLexemes(String inputString)
            throws UnitParserException {
        UnitParser p = new UnitParser(Syntax.FITS, inputString); // the parser type doesn't matter
        UnitParser.Lexeme lexeme;
        while ((lexeme = p.getLexeme()) != null) {
            System.out.println("Token: " + lexeme.token + ", yylval=" + lexeme.value);
        }
    }

    /**
     * Display the parsed version of the input string in an unambiguous way
     */
    private static void displayParse(String inputString, Syntax parser) {
        try {
            UnitExpr units = new UnitParser(parser, inputString).getParsed();
            System.out.print(units + ":");
            for (OneUnit u : units) {
                System.out.print("  " + u.toDebugString());
            }
            System.out.println();
        } catch (UnitParserException e) {
            System.err.println("Failed to parse that: " + e);
        }
    }

    /**
     * Run the tests in the named CSV file
     */
    private static void runTestsInFile(String filename, Syntax parser)
            throws UnitParserException {
        try {
            java.io.BufferedReader in =new java.io.BufferedReader(new java.io.FileReader(filename));
            String line;
            int nok = 0;
            int nfail = 0;

            // Set the locale to something non-default, so we can
            // check that the testing code is locale-independent (ie
            // that it doesn't just work by accident in a particular
            // locale).  The library as a whole should be
            // locale-independent.  This isn't
            // a very thorough test, because the French locale seems
            // to differ from the non-locale only in the use of commas
            // for the decimal separator.  However I can't find a
            // locale which differs more than this in relevant
            // respects -- all the ones I would have expected to be
            // different (japaense, chinese, arabic, thai, ...) seem to have
            // 'us' digits (pragmatically, I suppose).
            java.util.Locale.setDefault(new java.util.Locale("FR"));

            // Create a 'sloppy' CSV reader (ie it skips blank and '#'
            // lines, and skips whitespace after CSV commas)
            for (List<String> fields : new SimpleCSVReader(in, true)) {
                
                String unitString;
                boolean shouldBeOK;
                String failureReason;
                String parseResult;
                String expected = null;
                UnitExpr expr;

                // expectedStatus indicates further tests:
                final int STATUS_FINE = 0;               // no further tests
                final int STATUS_UNRECOGNISED_UNITS = 1; // includes unrecognised units
                final int STATUS_DEPRECATED_UNIT = 2;    // all units recognised, but at least one deprecated
                final int STATUS_FORBIDDEN_SI = 3;       // uses SI prefixes where forbidden
                final int STATUS_FORBIDDEN_BINARY = 4;   // uses binary prefixes where forbidden
                final int STATUS_ALL_CONSTRAINTS_OK = 5; // all constraints satisfied
                int expectedStatus = 0;

                // column 0
                {
                    unitString = fields.get(0);
                    StringBuffer sb = new StringBuffer();
                    try {
                        expr = new UnitParser(parser, fields.get(0)).getParsed();
                        parseResult = expr.toString(Syntax.DEBUG);
                    } catch (UnitParserException e) {
                        // that's OK -- silently acknowledge this
                        parseResult = null;
                        expr = null;
                    }
                }

                // column 1
                shouldBeOK = (fields.get(1).length() == 0);
                failureReason = fields.get(1);

                // column 2
                if (fields.size() > 2) {
                    char leadingChar = fields.get(2).charAt(0);
                    if (Character.isDigit(leadingChar) || leadingChar == '-') {
                        // OK
                        expectedStatus = STATUS_FINE;
                        expected = fields.get(2);
                    } else {
                        expected = fields.get(2).substring(1);
                        switch (leadingChar) {
                          case 'x':
                            expectedStatus = STATUS_UNRECOGNISED_UNITS;
                            break;
                          case '?':
                            expectedStatus = STATUS_DEPRECATED_UNIT;
                            break;
                          case 's':
                            expectedStatus = STATUS_FORBIDDEN_SI;
                            break;
                          case 'b':
                            expectedStatus = STATUS_FORBIDDEN_BINARY;
                            break;
                          case '!':
                            expectedStatus = STATUS_ALL_CONSTRAINTS_OK;
                            break;
                          default:
                            // oops -- the above are the only allowed flags
                            System.err.println("Malformed test: unrecognised flag '" + leadingChar + "' in " + fields.get(2));
                            System.exit(1);
                        }
                    }
                }

                // we've gathered all the information -- now do the check

                if (shouldBeOK) {
                    if (parseResult == null) {
                        System.out.println("Parsing \"" + unitString
                                           + "\" should have worked, but didn't");
                        nfail++;
                    } else if (parseResult.equals(expected)) {
                        // this test passed
                        nok++;
                        switch (expectedStatus) {
                          case STATUS_FINE:
                            break; // no other checks required
                          case STATUS_UNRECOGNISED_UNITS:
                            // this string had non-recognised units
                            if (expr.allUnitsRecognised(parser)) {
                                System.out.println("Parsing \"" + unitString
                                                   + "\"(" + parser + ") has non-recognised units: didn't spot them!");
                                nfail++;
                            } else {
                                nok++;
                            }
                            break;
                          case STATUS_DEPRECATED_UNIT:
                            // all units recognised, but at least one deprecated
                            if (expr.allUnitsRecognised(parser) && !expr.allUnitsRecommended(parser)) {
                                nok++;
                            } else {
                                System.out.println("Parsing \"" + unitString
                                                   + "\"(" + parser + ") has deprecated units: didn't spot them");
                                nfail++;
                            }
                            break;
                          case STATUS_FORBIDDEN_SI:
                          case STATUS_FORBIDDEN_BINARY:
                            // at least one unit was used with an SI
                            // prefix.  The unit was recognised but is
                            // known not to permit this.  We don't (I
                            // think) have to distinguish these two
                            // cases, since they're bot aliases for
                            // the same underlying check:
                            if (expr.allUsageConstraintsSatisfied(parser)) {
                                System.out.println("Parsing \"" + unitString
                                                   + "\"(" + parser + ") does not conform to usage constraints: didn't spot that");
                                nfail++;
                            } else {
                                nok++;
                            }
                            break;
                          case STATUS_ALL_CONSTRAINTS_OK:
                            // fully conformant
                            if (expr.isFullyConformant(parser)) {
                                nok++;
                            } else {
                                System.out.println("Parsing \"" + unitString
                                                   + "\"(" + parser + ") is fully conformant: didn't spot that");
                                nfail++;
                            }
                            break;
                          default:
                            throw new AssertionError("Impossible expectedStatus: " + expectedStatus);
                        }
                    } else {
                        System.out.println("Parsing \"" + unitString
                                           + "\": expected \""  + expected
                                           + "\" got \"" + parseResult + "\"");
                        nfail++;
                    }
                    // separately, check the formatted versions, if they're present
                    if (expr != null && fields.size() > 3) {
                        // FIXME: should check that reparsing the formatted produces the same expr
                        // (define UnitExpr.equals(UnitExpr) method?)
                        for (int i=3; i<fields.size(); i++) {
                            String[] parts = fields.get(i).split(":");
                            if (parts.length != 2) {
                                System.err.print("Malformed formatting specifier: "
                                                 + fields.get(i) + " on line: ");
                                for (int ii=0; ii<fields.size(); ii++) {
                                    System.err.print(fields.get(ii) + ", ");
                                }
                                System.err.println();
                            } else {
                                Syntax syntax = Syntax.lookup(parts[0]);
                                if (syntax == null) {
                                    System.err.println("Bad test: field "
                                                       + fields.get(i)
                                                       + " mentions invalid parser");
                                    System.exit(1);
                                }
                                String actualFormat = expr.toString(syntax);
                                String expectedFormat = parts[1];
                                if (actualFormat.equals(expectedFormat)) {
                                    nok++;
                                } else {
                                    System.out.println("Formatting (" + syntax
                                                       + "): expected \"" + expectedFormat
                                                       + "\", got \"" + actualFormat + "\"");
                                    nfail++;
                                }

                                // now re-parse the result
                                //System.out.println("Re-parsing " + actualFormat + " with syntax " + parts[0]);
                                UnitParser p;
                                try {
                                    p = new UnitParser(syntax, actualFormat);
                                } catch (UnitParserException e) {
                                    // that's OK: this is a write-only syntax
                                    p = null;
                                }
                                if (p != null) {
                                    try {
                                        UnitExpr reExpr = p.getParsed();
                                        String reParseResult = reExpr.toString(Syntax.DEBUG);
                                        if (! reParseResult.equals(expected)) {
                                            System.out.println("Re-parsing " + actualFormat + " with syntax " + syntax
                                                               + ": produced " + reParseResult + ", expected " + expected);
                                            nfail++;
                                        }
                                    } catch (UnitParserException e) {
                                        // parsing failed
                                        System.out.println("Re-parsing " + actualFormat + " with syntax " + syntax
                                                           + ": FAILED [" + e + "]");
                                        nfail++;
                                    }
                                }
                            }
                        }
                    }
                } else {
                    // should have failed
                    if (parseResult == null) {
                        // this test was expected to fail
                        nok++;
                    } else {
                        System.out.println("Parsing \"" + unitString
                                           + "\" should have failed ("
                                           + failureReason
                                           + "), but it produced: "
                                           + parseResult);
                        nfail++;
                    }
                }
            }
            StringBuilder sb = new StringBuilder("        ");
            String syntaxName = parser.toString();
            sb.replace(sb.length()-syntaxName.length(), sb.length(), syntaxName);
            if (nfail == 0) {
                System.out.println(sb + " : ALL " + nok + " tests pass");
            } else {
                System.out.println(sb + " : passes: " + nok);
                System.out.println("           FAILS:  " + nfail);
            }
        } catch (java.io.IOException e) {
            System.err.println("IOException reading " + filename + ": " + e);
        }
    }
    
    private static void Usage() {
        System.err.println("Usage: ParseAll -lpt unitstring");
        System.exit(1);
    }
}
