package uk.me.nxg.unity;

import java.util.List;
import java.util.regex.Pattern;
import java.util.regex.Matcher;


/**
 * A single simple unit, such as 'kg'.
 * This is represented as a <em>prefix</em>,
 * a <em>base unit</em> and an <em>exponent</em>.  For example, the
 * string "mm s^-2" will result in two SimpleUnit instances, one of which
 * will have a base unit of "m" and a prefix power of -3, and the
 * other of which will have a base unit of "s" and an exponent of -2.
 *
 * <p>This represents the abstract unit, shorn of the symbols which
 * were parsed to obtain it.  Thus if, in some odd syntax, the symbol
 * 'x' were used to denote metres, then the SimpleUnit instance which
 * resulted would be the same as the OneUnit which resulted from a
 * more conventional syntax. 
 *
 * <p>Also, there is a potential ambiguity if a symbol is recognised
 * in one syntax, but not in another.  Thus if the string 'erg' were
 * parsed in a syntax which didn't recognise that, then it would be
 * stored as just that, an unrecognised symbol, not associated with
 * the {@link UnitDefinition} for the erg.
 *
 * <p>The class's instances are immutable.
 */
public abstract class SimpleUnit extends OneUnit {
    private final UnitDefinitionMap.Resolver unitResolver;

    /**
     * Represents the power-of-ten (or power-of-two) 
     * modifier in front of the base unit.
     * For example, in the string "mm", the prefix power would be "-3".
     */
    private final int prefixPower;

    /**
     * The base unit as a UnitDefinition.
     * The 'base unit' is the unit without the prefix.
     * Thus in the expression "MW", it is 'W', Watt, that is the base unit.
     * This doesn't refer to the fundamental SI base units.
     * Precisely one of baseUnitString and baseUnitDefinition is non-null.
     */
    private final UnitDefinition baseUnitDefinition;

    /**
     * The base unit as a string.  If this is non-null, it means that
     * the unit string wasn't recognised as a known-unit.
     */
    private final String baseUnitString;

    /**
     * The power to which this base unit is raised.
     */
    //private final float exponent;

    /**
     * We store the resolver which was used to look up the symbols
     * which identified this unit.  This makes it possible to pass
     * this along to the UnitExpr class.
     */
    //private final UnitDefinitionMap.Resolver unitResolver;

    /** 
     * A regexp to match a unit with an SI prefix.
     * The order in this regexp is important -- the "da" must match
     * before "d" can, so that "dam" is parsed as "da"+"m", and not
     * "d"+"am".  Also, this regexp requires there to be a unit: "m"
     * is not parsed as a prefix plus an empty unit.
     */
    // private static final Pattern unitWithPrefix
    //         = Pattern.compile("(da|[yzafpnumcdhkMGTPEZY])(.+)");

    /**
     * A regexp which matches a unit with an ISO/IEC 80000-13 binary prefix.
     */
    // private static final Pattern unitWithBinaryPrefix
    //         = Pattern.compile("([KMGTPEZY]i)(.+)");

    /**
     * Constructor is intended to be internal, but has to be invokable
     * by subclasses.
     */
    SimpleUnit(UnitDefinitionMap.Resolver r, int prefix, String us, UnitDefinition ud, float e) {
        super(e);
        prefixPower = prefix;
        if (us == null) {
            baseUnitString = null;
            baseUnitDefinition = ud;
        } else {
            baseUnitString = us;
            baseUnitDefinition = null;
        }
        //exponent = e;

        // class invariants
        assert    (baseUnitString == null && baseUnitDefinition != null)
               || (baseUnitString != null && baseUnitDefinition == null);
        this.unitResolver = r;
        assert unitResolver != null;
    }

    // static SimpleUnit makeSimpleUnit(UnitDefinitionMap.Resolver r,
    //                                  String unitString,
    //                                  int e) {
    //     return makeSimpleUnit(r, unitString, (float)e);
    // }
    
    static SimpleUnit makeSimpleUnit(UnitDefinitionMap.Resolver r,
                                     String unitString,
                                     float e) {
                // It would be possible/appropriate to do strict checking of
        // units in here -- that is, throwing exceptions whenever we
        // encounter any non-recommended unit (ie one which lookupUnit
        // returns null for).  This would create the additional
        // possibility of being a convenient/reasonable place to check
        // whether the abbreviation that we look up is an allowed one
        // for the syntax we're parsing, as long as we were
        // additionally provided with some information within this
        // constructor, about which syntax that is (which is currently
        // not available in here).
        UnitDefinition def = r.lookupSymbol(unitString);

        SimpleUnit rval;
        
        if (def != null) {
            rval = new SimpleDecimalUnit(r, 0, null, def, e);
        } else {
            PrefixSplit ps = SimpleBinaryUnit.splitUnitString(unitString);
            if (ps != null) {
                String us = ps.getUnit();
                def = r.lookupSymbol(us);
                if (def == null) {
                    rval = new SimpleBinaryUnit(r, ps.getPower(), us, null, e);
                } else {
                    rval = new SimpleBinaryUnit(r, ps.getPower(), null, def, e);
                }
            } else {
                ps = SimpleDecimalUnit.splitUnitString(unitString);
                if (ps != null) {
                    String us = ps.getUnit();
                    def = r.lookupSymbol(us);
                    if (def == null) {
                        rval = new SimpleDecimalUnit(r, ps.getPower(), us, null, e);
                    } else {
                        rval = new SimpleDecimalUnit(r, ps.getPower(), null, def, e);
                    }
                } else {
                    rval = new SimpleDecimalUnit(r, 0, unitString, null, e);
                }
            }
        }

        assert rval != null;
        return rval;
    }

    // /**
    //  * Constructor of base unit -- no multiple.
    //  */
    // private SimpleUnit(UnitDefinitionMap.Resolver r, String s, int e) {
    //     this(r, s, (float)e);
    // }

    // /**
    //  * Constructor of base unit
    //  * @param unitString a string representation of a single unit with
    //  * a possible SI prefix
    //  * @param e the power to which this unit is raised
    //  */
    // private SimpleUnitXXX(UnitDefinitionMap.Resolver r, String unitString, float e) {
    //     //super(e);
    //     // this.unitResolver = r;
    //     // assert unitResolver != null;

    //     // It would be possible/appropriate to do strict checking of
    //     // units in here -- that is, throwing exceptions whenever we
    //     // encounter any non-recommended unit (ie one which lookupUnit
    //     // returns null for).  This would create the additional
    //     // possibility of being a convenient/reasonable place to check
    //     // whether the abbreviation that we look up is an allowed one
    //     // for the syntax we're parsing, as long as we were
    //     // additionally provided with some information within this
    //     // constructor, about which syntax that is (which is currently
    //     // not available in here).
    //     UnitDefinition def = unitResolver.lookupSymbol(unitString);

    //     if (def != null) {
    //         prefixPower = 0;
    //         baseUnitString = null;
    //         baseUnitDefinition = def;
    //     } else {
    //         java.util.regex.Matcher m = unitWithPrefix.matcher(unitString);
    //         if (m.matches()) {
    //             prefixPower = prefixLetterToPower(m.group(1).equals("da")
    //                                               ? UnitVal.CHAR_DECA
    //                                               : m.group(1).charAt(0));
    //             def = unitResolver.lookupSymbol(m.group(2));
    //             if (def == null) {
    //                 baseUnitString = m.group(2);
    //                 baseUnitDefinition = null;
    //             } else {
    //                 baseUnitString = null;
    //                 baseUnitDefinition = def;
    //             }
    //         } else {
    //             prefixPower = 0;
    //             baseUnitString = unitString;
    //             baseUnitDefinition = null;
    //         }
    //     }
    //     //exponent = e;

    //     // class invariant
    //     assert (baseUnitString == null && baseUnitDefinition != null) || (baseUnitString != null && baseUnitDefinition == null);
    // }

    /**
     * Produce the reciprocal of the unit, which is a unit which has
     * the negative of the exponent of this one.
     * @return a new instance
     */
    abstract SimpleUnit reciprocate();
    
    // SimpleUnit reciprocate() {
    //     return new SimpleUnit(unitResolver, prefixPower, baseUnitString, baseUnitDefinition, -getExponent());
    // }

    /**
     * Reciprocate a list of units
     */
    /*
    static List<SimpleUnit> reciprocate(Iterable<SimpleUnit> ul) {
        List<SimpleUnit> res = new java.util.ArrayList<OneUnit>();
        for (OneUnit u : ul) {
            res.add(u.reciprocate());
        }
        return res;
    }
    */
    /**
     * Divide one list of units by another
     */
    /*
    static List<OneUnit> divide(List<OneUnit> num, Iterable<OneUnit> den) {
        List<OneUnit> res = new java.util.ArrayList<OneUnit>(num);
        for (OneUnit u : den) {
            res.add(u.reciprocate());
        }
        return res;
    }
    */

    /**
     * Returns the prefix of the unit, as a base-ten log.  Thus a
     * prefix of "m", for "milli", would produce a prefix of -3.
     */
    @Override public int getPrefix() {
        return prefixPower;
    }

    // /**
    //  * Returns the power to which this unit is raised.
    //  * A unit of "mm^3" would return 3.0, and "m^(-1/2)" would return -0.5.
    //  */
    // public float getPower() {
    //     return exponent;
    // }
    
    // private static int prefixLetterToPower(char prefixLetter) {
    //     int prefixPower;
    //     switch (prefixLetter) {
    //       case 'y': prefixPower = -24; break;
    //       case 'z': prefixPower = -21; break;
    //       case 'a': prefixPower = -18; break;
    //       case 'f': prefixPower = -15; break;
    //       case 'p': prefixPower = -12; break;
    //       case 'n': prefixPower =  -9; break;
    //       case 'u': prefixPower =  -6; break;
    //       case 'm': prefixPower =  -3; break;
    //       case 'c': prefixPower =  -2; break;
    //       case 'd': prefixPower =  -1; break;
    //       case UnitVal.CHAR_DECA:
    //         prefixPower = 1;
    //         break;
    //       case 'h': prefixPower =   2; break;
    //       case 'k': prefixPower =   3; break;
    //       case 'M': prefixPower =   6; break;
    //       case 'G': prefixPower =   9; break;
    //       case 'T': prefixPower =  12; break;
    //       case 'P': prefixPower =  15; break;
    //       case 'E': prefixPower =  18; break;
    //       case 'Z': prefixPower =  21; break;
    //       case 'Y': prefixPower =  24; break;
    //       default:
    //         // The lexer in lex.jflex should make this impossible
    //         throw new AssertionError("Impossible prefix: '" + prefixLetter + "'");
    //     }
    //     return prefixPower;
    // }

    /**
     * Convert a prefix power to a String.  These are returned as
     * Strings, so that we can later consider 'da' or 'ki' (etc) as prefixes.
     */
    abstract String prefixPowerToString();
    // private static String prefixPowerToString(int prefixPower) {
    //     String rval;
    //     switch (prefixPower) {
    //       case -24: rval = "y"; break;
    //       case -21: rval = "z"; break;
    //       case -18: rval = "a"; break;
    //       case -15: rval = "f"; break;
    //       case -12: rval = "p"; break;
    //       case  -9: rval = "n"; break;
    //       case  -6: rval = "u"; break;
    //       case  -3: rval = "m"; break;
    //       case  -2: rval = "c"; break;
    //       case  -1: rval = "d"; break;
    //       case   0: rval = "";  break;
    //       case   1: rval = "da";break;
    //       case   2: rval = "h"; break;
    //       case   3: rval = "k"; break;
    //       case   6: rval = "M"; break;
    //       case   9: rval = "G"; break;
    //       case  12: rval = "T"; break;
    //       case  15: rval = "P"; break;
    //       case  18: rval = "E"; break;
    //       case  21: rval = "Z"; break;
    //       case  24: rval = "Y"; break;
    //       default:
    //         // This should be impossible, since it's only
    //         // prefixLetterToPower which sets these values
    //         throw new AssertionError("Impossible prefix power: " + prefixPower);
    //     }
    //     return rval;
    // }

    /**
     * Convert a prefix power to a String, in a form useful within
     * siunitx.  The only real difference from
     * {@link prefixPowerToString}
     * is that <code>\micro</code> is formatted differently.
     */
    abstract String prefixPowerToStringLaTeX();
    // private static String prefixPowerToStringLaTeX(int prefixPower) {
    //     String rval;
    //     switch (prefixPower) {
    //       case -24: rval = "\\yocto"; break;
    //       case -21: rval = "\\zepto"; break;
    //       case -18: rval = "\\atto";  break;
    //       case -15: rval = "\\femto"; break;
    //       case -12: rval = "\\pico";  break;
    //       case  -9: rval = "\\nano";  break;
    //       case  -6: rval = "\\micro"; break;
    //       case  -3: rval = "\\milli"; break;
    //       case  -2: rval = "\\centi"; break;
    //       case  -1: rval = "\\deci";  break;
    //       case   0: rval = "";        break;
    //       case   1: rval = "\\deca";  break;
    //       case   2: rval = "\\hecto"; break;
    //       case   3: rval = "\\kilo";  break;
    //       case   6: rval = "\\mega";  break;
    //       case   9: rval = "\\giga";  break;
    //       case  12: rval = "\\tera";  break;
    //       case  15: rval = "\\peta";  break;
    //       case  18: rval = "\\exa";   break;
    //       case  21: rval = "\\zetta"; break;
    //       case  24: rval = "\\yotta"; break;
    //       default:
    //         // This should be impossible, since it's only
    //         // prefixLetterToPower which sets these values
    //         throw new AssertionError("Impossible prefix power: " + prefixPower);
    //     }
    //     return rval;
    // }

    /**
     * Given a unit string, detect whether there is an ISO/IEC
     * 80000-13 binary prefix at the front of it, and return a
     * positive integer if so.  If there is not, return a negative
     * integer.  These are all two characters long.
     */
    // private static int prefixToBinaryPower(String s) {
    //     int prefixPower;
    //     java.util.regex.Matcher m = binaryPrefix.matcher(s);
    //     if (m.matches()) {
    //         switch (s.charAt(0)) {
    //           case 'K': prefixPower = 10; break;
    //           case 'M': prefixPower = 20; break;
    //           case 'G': prefixPower = 30; break;
    //           case 'T': prefixPower = 40; break;
    //           case 'P': prefixPower = 50; break;
    //           case 'E': prefixPower = 60; break;
    //           case 'Z': prefixPower = 70; break;
    //           case 'Y': prefixPower = 80; break;
    //           default:
    //             assert false;
    //         }
    //         return prefixPower;
    //     } else {
    //         return -1;
    //     }
    // }
    

    @Override public String toString() {
        return toString(null);
    }
    
    @Override public String toString(Syntax syntax) {
        StringBuilder sb = new StringBuilder();
        try {
            sb.append(unitString(syntax));
        } catch (UnitParserException e1) {
            try {
                sb.append(unitString(Syntax.FITS));
            } catch (Exception e2) {
                // this should be impossible
                throw new AssertionError("SimpleUnit.toString with syntax FITS really should work: " + e2);
            }
        }
        if (getExponent() != 1) {
            sb.append('^').append(getExponent());
        }
        return sb.toString();
    }

    public String unitString(Syntax syntax)
            throws UnitParserException {
        String ret;
        if (syntax == Syntax.LATEX) {
            ret = unitStringLaTeX();
        } else {
            ret = unitStringText(syntax);
        }
        assert ret != null;
        return ret;
    }

    @Override public UnitDefinition getBaseUnitDefinition() {
        return baseUnitDefinition;
    }

    @Override public String getBaseUnitName() {
        assert baseUnitString != null || baseUnitDefinition != null;
        return (baseUnitDefinition != null ? baseUnitDefinition.name() : baseUnitString);
    }

    @Override public String getBaseUnitString() {
        return baseUnitString;
    }

    @Override public Dimensions getDimensions() {
        if (baseUnitDefinition == null) {
            return null;
        } else {
            return baseUnitDefinition.dimensions();
        }
    }
    
    /**
     * Retrieve the string form of the unit in one of the text syntaxes.
     * @param syntaxName one of the syntaxes of {@link Syntax}
     * @return the text unit as, for example, "mm" for millimetres
     * @throws UnitParserException if the syntax is unrecognised
     */
    private String unitStringText(Syntax syntax)
            throws UnitParserException {
        StringBuilder sb = new StringBuilder();
        if (prefixPower != 0) {
            sb.append(prefixPowerToString());
        }
        if (baseUnitString != null) {
            sb.append(baseUnitString);
        } else {
            UnitDefinitionMap.Resolver r;
            if (syntax == null) {
                // special call, from unitString(void)
                r = unitResolver;
            } else {
                r = UnitDefinitionMap.getInstance().getResolver(syntax);
                if (r == null) {
                    throw new UnitParserException("Unknown syntax: " + syntax);
                }
            }
            assert baseUnitDefinition != null;
            UnitRepresentation rep = r.lookupUnit(baseUnitDefinition);
            if (rep == null) {
                // this UnitDefinition is not recognised in this syntax
                rep = unitResolver.lookupUnit(baseUnitDefinition);
                // this resolver is where we originally got the
                // baseUnitDefinition from, so this must succeed
            }
            assert rep != null;
            sb.append(rep.getAbbreviation());
        }
        return sb.toString();
    }

    /**
     * Retrieve the string form of the unit in a form suitable for use
     * within the siunitx LaTeX package.  The LaTeX serialisation
     * isn't quite a syntax like the text ones, though it appears
     * similar to one from the outside.
     * @return the text unit as, for example, "mm" for millimetres
     * @throws UnitParserException if the syntax is unrecognised
     */
    private String unitStringLaTeX()
            throws UnitParserException {
        StringBuilder sb = new StringBuilder();
        if (prefixPower != 0) {
            sb.append(prefixPowerToStringLaTeX()).append(' ');
        }
        if (baseUnitString != null) {
            sb.append(baseUnitString);
        } else {
            assert baseUnitDefinition != null; // from class invariant

            if (baseUnitDefinition.latexForm() != null) {
                sb.append(baseUnitDefinition.latexForm());
            } else {
                UnitRepresentation rep = unitResolver.lookupUnit(baseUnitDefinition);
                // this unitResolver is where we originally got the
                // baseUnitDefinition from, so this must succeed
                assert rep != null;

                sb.append(rep.getAbbreviation());
            }
        }
        return sb.toString();
    }
    
    /**
     * Obtains the string form of the unit, including prefix, with a
     * default syntax.  That is, return 'mm' not 'm'.
     *
     * <p>The default syntax is (currently) the syntax with which this
     * unit was originally read.
     */
    public String unitString() {
        try {
            return unitString(null);
        } catch (UnitParserException e) {
            throw new AssertionError("Odd: SimpleUnit.unitString() should not have failed");
        }
    }
    
    @Override UnitDefinitionMap.Resolver getUnitResolver() {
        return unitResolver;
    }

    @Override public boolean isRecognisedUnit(Syntax syntax) {
        // We could check the validity of the abbreviation whilst we're
        // parsing the units, but that would frustrate our ability to
        // subsequently write out recommended units in the appropriate
        // form.
        return baseUnitDefinition != null
                && baseUnitDefinition.getRepresentation(syntax) != null;
    }

    @Override public boolean isRecommendedUnit(Syntax syntax) {
        // We could check the validity of the abbreviation, when we're
        // parsing the units, but that would frustrate our ability to
        // subsequently write out recommended units in the appropriate
        // form.
        if (baseUnitDefinition == null) {
            return false;
        } else {
            UnitRepresentation r = baseUnitDefinition.getRepresentation(syntax);
            return r != null && !r.isDeprecated();
        }
    }

    /**
     * Indicate whether this unit has permitted prefixes.  The only
     * constraints are for 'known' units, so we only check cases which
     * have a UnitRepresentation associated with them.
     * @return true if this unit has an acceptable prefix
     */
    abstract boolean hasPermittedPrefix(UnitRepresentation r);

    @Override public boolean satisfiesUsageConstraints(Syntax syntax) {
        // Other constraints are possible: for some syntaxes, the 'Crab'
        // unit may be used only with the 'm' prefix.
        boolean validity;
        if (baseUnitDefinition == null) {
            // there are no constraints, so this necessarily satisfies all of them
            validity = true;
        } else {
            UnitRepresentation r = baseUnitDefinition.getRepresentation(syntax);
            if (r == null) {
                // as above
                validity = true;
            } else {
                validity = hasPermittedPrefix(r);
                //validity = prefixPower == 0 || r.mayHaveSIPrefixes();
            }
        }
        return validity;
    }

    @Override public String toDebugString() {
        StringBuilder sb = new StringBuilder();
        sb.append(prefixPower);
        if (this instanceof SimpleBinaryUnit) {
            sb.append('b');
        }
        sb.append('/');
        if (baseUnitString == null) {
            // first try this with a given syntax, for the sake of
            // consistency, but fall back to an any-syntax option
            // without qualms
            UnitRepresentation r = baseUnitDefinition.getRepresentation(Syntax.FITS);
            if (r == null) {
                // fallback
                r = baseUnitDefinition.getRepresentation();
            }
            sb.append(r.getAbbreviation());
        } else {
            sb.append(baseUnitString);
        }
        sb.append('/');
        float e = getExponent();
        if (e == Math.floor(e)) { // no fractional part
            sb.append(Math.round(e)); // print an integer
        } else {
            sb.append(e);
        }
        return sb.toString();
    }

    /**
     * Two units are equal if they have the same power, units and
     * exponent.  They don't have to have been obtained from the same
     * syntax, so that in a syntax which (for example) permitted both
     * "yr" and "a" for years, two SimpleUnit instances, obtained from
     * parsing the two alternatives, would be regarded as equal.
     */
    @Override public boolean equals(Object o) {
        // Implementation note: the practical effect of the above
        // notes is that the 'resolver' field should be ignored for
        // the purposes of the equals and hashCode methods.
        if (o == this) {
            return true;        // trivially
        } else if (o instanceof SimpleUnit) {
            SimpleUnit oo = (SimpleUnit)o;
            return oo.prefixPower == prefixPower
                    && oo.baseUnitDefinition == baseUnitDefinition
                    && oo.baseUnitString == baseUnitString
                    && oo.getExponent() == getExponent();
        } else {
            return false;
        }
    }

    @Override public int hashCode() {
        // We've overridden equals(), so we have to override hashCode,
        // too (as noted in eg Bloch, Effective Java, prescription 8,
        // and using the simple approach recommended there).
        int result = 17;
        result = 37*result + prefixPower;
        if (baseUnitDefinition == null) {
            assert baseUnitString != null;
            result = 37*result + baseUnitString.hashCode();
        } else {
            assert baseUnitString == null;
            result = 37*result + baseUnitDefinition.hashCode();
        }
        long de = Double.doubleToLongBits(getExponent());
        result = 37*result + (int)(de ^ (de >>> 32));
        return result;
    }

    interface PrefixSplit {
        public int getPower();
        public String getUnit();
    }
}
